#ifndef __MMC_PROTOCOLS_H
#define __MMC_PROTOCOLS_H

/*
       MMC (My Mail Client). The name says it all. A Gnome Mail Client.
       Copyright (C) 2000  Christian Biesinger <biesi@users.sourceforge.net> and
       Steven Behnke <sbehnke@digistatic.net>

       This program is free software; you can redistribute it and/or modify
       it under the terms of the GNU General Public License as published by
       the Free Software Foundation; either version 2 of the License, or
       (at your option) any later version.

       This program is distributed in the hope that it will be useful,
       but WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
       GNU General Public License for more details.

       You should have received a copy of the GNU General Public License
       along with this program; if not, write to the Free Software
       Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

       For the complete License see the file COPYING.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <gtk/gtk.h>
#include <sys/socket.h>
#include <unistd.h>

extern gboolean isqp, fl;

gchar *oneLine(const gchar *lines, const gchar *sep="\n");
// Returns: Newly allocated string. delete[] it.
// it makes from "Line1\nLine2" "Line1\n"

gchar *getHead(const gchar *message, const gchar *header);
// Message: contains headers and body
// header is the header to look for, including colon ":"
// Result must be deleted, and is for example:
// "From: Christian Biesinger <biesi@users.sourceforge.net>"
// Returns NULL if header not found.

gchar *dup(const gchar *string);
// Works like strdup, but result should be
// delete[]ed not free()d or g_free()d

void append(gpointer data1, gpointer data2);
void appendgl(gpointer, gpointer);

enum PROTO_FLAGS {
	FLAG_SEND = 1 << 1,
	FLAG_RECV = 1 << 2
};

enum FOLD_FLAGS {
	FOLD_ADD    = 1 << 1,
	FOLD_CREATE = 1 << 2
};

enum PROTO_ERR {
	OK,        // Everything was ok
	ERR_CONN,  // Connection failed
	ERR_AUTH,  // Authentication failed
	ERR_RETR,  // Message Retrieving failed
	ERR_OTHER  // Something else failed
	
};

struct listflag {
	GList* list;
	int flags;
};

class CQueryable {
public:
	virtual char* query() const=0;
	// NEVER free the result

	// following returns PROTO_SEND or PROTO_RECV or
	// PROTO_SEND|PROTO_RECV
	virtual gint flags() = 0;
};


// Derived Classes that should be there:
// POP3
// IMAP
// POP3, SOCKS
// IMAP, SOCKS
// SMTP
// maybe local Sendmail
//
// Note: if only send or only recv plugin:
// overload unused functions with a function
// that doesn't do anything:
// e.g.: int CProtoSMTP::get() {return 0;}
class CProto : public CQueryable {
public:
	//the following HAS a reason
	int init(GtkCList* l, GtkText* t) {return 0;}

	static void error(PROTO_FLAGS flag, const gchar* where, const gchar* line);
	
	// Retrieve Mails from a server
	virtual int get(const gchar *data)=0;

	// Send a mail
	virtual int send(const gchar *string, const gchar *from, gchar *to[],
		const gchar *data, GtkWidget *dialog)=0;
	// from,to,data are clear
	// dialog is the dialog (a GnomeApp or
	// a GtkWindow) or NULL.
	// If not NULL, it should be destroyed after the
	// mail is sent.
	// string is the string given back in applyConfig
	// must return 0 on success
	
	virtual GtkWidget *config(PROTO_FLAGS flag, const gchar *account, const gchar *yourstring, GtkWidget *prefs)=0;
	virtual gchar* applyConfig(PROTO_FLAGS flag, GtkWidget* widget)=0;
	virtual void free(gchar* str) { delete[] str; }
	// YOU MUST NAME THE LAST ARGUMENT PREFS!
	// (if you will use the CAN_CHANGE macro)
	
	// They return a string.
	// The functions get() and send() get passed this string,
	// so that they can find their coniguration
	// config gets the account-description as the
	// second parameter
	//
	// free is used to free what is returned by applyConfig
	
	virtual int conn(const char* hostname, int port); // returns 0 on error
	virtual int conn(guint32 ip, int port);
	// Connect, return socknum (usually)
	// Are defined in CProto.
	// but can be overwritten in derived classes:
	// useful for SOCKS support or SSL Support
	virtual int send(int fd, const void* data, size_t length, int flags) {
		return ::send(fd, data, length, flags);
	}
	virtual int recv(int fd, void* data, size_t len, int flags) {
		return ::recv(fd, data, len, flags);
	}
	virtual int close(int fd) { return ::close(fd); }

	virtual ssize_t recvline(int sock, char *vptr, size_t maxlen);
	


protected:
};

// We should have:
// IMAP
// MBOX
class CFolder : public CQueryable {
public:
	virtual ~CFolder() { return; }

	virtual int init(GtkCList *mails, GtkText *text)=0;
	// Reads the list of configured Folders
	// Does g_folders.add for each one
	
	virtual int create()=0;
	// Creates a folder. Usually Displays a dialog and then
	// creates the folder.
	
	virtual int addFolder()=0;
	// Adds an existing folder. Also displays a dialog.
	// Not all plugins support both options, I assume - I'm not
	// sure if IMAP supports creation of folders.
	// Now I know. IMAP does.

	//==============================================================
	// Above this Line: functions should be static; unfortunately,
	// they can't be because they must be virtual.
	// This means that they apply for all folders of this type
	// Below: Apply for a single folder.
	//==============================================================

	virtual int remove(gboolean onlyfromlist)=0;
	// onlyfromlist is true when the folder should only
	// be removed from the folder list

	virtual void free(void*)=0;
	// Call this to free the data stored in CMessage.
	// Probably only used by CMessage

	virtual int parse()=0;
	// Adds the mails to the CList mail_list.
	// It has 3 Columns: Subject, Date, From (in this order).

	virtual int select(guint row) { return select(get(row), fl, isqp, row, gtk_clist_get_row_data(mail_list, row)); }
	virtual int select(void* data)=0;
	virtual int select(const gchar *text, gboolean all_headers, gboolean isqp,
		gint row, void* data);
	// displays the mail
	
	virtual int addMail(const gchar *raw)=0; // permanently
	virtual int addMail(const gchar* from, const gchar *subject,
 			const gchar *date, void *data); // not permanently
	// adds a mail to the folder.
	// creates a CMessage* sets this as the row data
	
	virtual char* get(gint row)=0;
	virtual char* get(void* data)=0;
	// Gets the raw message.
	
	virtual int del(void* data)=0; // deletes a message from list&file
	virtual int del(gint row)=0;

	virtual void markRead(void* data)=0;
	// Marks the Message as read

	// -------------------------------------------------
	// Networking functions
	// -------------------------------------------------
	virtual int conn(const char* hostname, int port); // returns 0 on error
	virtual int conn(guint32 ip, int port);
	// Connect, return socknum (usually)
	// Are defined in CFolder.
	// but can be overwritten in derived classes:
	// useful for SOCKS support or SSL Support
	virtual int send(int fd, const void* data, size_t length, int flags) {
		return ::send(fd, data, length, flags);
	}
	virtual int recv(int fd, void* data, size_t len, int flags) {
		return ::recv(fd, data, len, flags);
	}
	virtual int close(int fd) { return ::close(fd); }
	virtual ssize_t recvline(int sock, char *vptr, size_t maxlen);

	int flags() { return FLAG_SEND|FLAG_RECV; }

	
protected:
	GtkText *text;
	GtkCList *mail_list;
};



template<class T> class Handler {
	public:
		Handler();

		//Adds a folder "plugin" to the internal list
		int add(T* fold, GtkCList* mlist, GtkText *te);
		int add(T* fold);

		T* get(const char *id);
		T* get(int id);
				
		GtkWidget *makeCList(gint flags);
		GtkWidget *makeCombo(gint flags);
	protected:
		GList *list;
		guint num;
};

template<class T> Handler<T>::Handler() {
	num=0;
	list=NULL;
}

template<class T> int Handler<T>::add(T *element) {
	list=g_list_append(list, element);
	num=g_list_length(list);
	return (num-1);
	
}

template<class T> int Handler<T>::add(T *element, GtkCList *mlist, GtkText *te) {
	element->init(mlist,te);
	return add(element);
}

template<class T> T* Handler<T>::get(const char* id) {
	char *q;

	if (num==0) return (T*)NULL;
	for (guint i=0; i<num; i++) {
		q=((CFolder*)g_list_nth_data(list,i))->query();

		if (strcasecmp(id,q)==0) {
			return (T*)g_list_nth_data(list,i);
		}
	}
	return NULL;
}

template<class T> T* Handler<T>::get(int id) {
	return (T*)(g_list_nth_data(list, id));
}

template<class T> GtkWidget* Handler<T>::makeCList(gint flags) {
#if 0
	GtkWidget *clist = gtk_clist_new(1);
	g_list_foreach(list, append, clist);
	gtk_widget_show(clist);
	return clist;
#endif
	GtkWidget* theList;
	gchar* strings[2];
	strings[1]=NULL;

	theList = gtk_clist_new(1);
	gtk_clist_set_selection_mode(GTK_CLIST(theList), GTK_SELECTION_SINGLE);


	T* fi;
	// GList* folders
	for (guint i = 0; i < g_list_length(list); i++) {
		fi = (T*)g_list_nth_data(list, i);
		if (!(fi->flags() & flags)) continue;
		strings[0] = fi->query();
		gtk_clist_append(GTK_CLIST(theList), strings);
	}

	gtk_clist_set_column_width(GTK_CLIST(theList), 0, 
		gtk_clist_optimal_column_width(GTK_CLIST(theList), 0));

	return theList;

}

template<class T> GtkWidget* Handler<T>::makeCombo(gint flags) {
	GtkWidget *ret = gtk_combo_new();
	listflag lf;
	lf.list = NULL;
	lf.flags = flags;
	g_list_foreach(list, appendgl, &lf);
	gtk_combo_set_popdown_strings(GTK_COMBO(ret), lf.list);
	gtk_combo_set_value_in_list(GTK_COMBO(ret), TRUE, FALSE);
	gtk_widget_show(ret);
	return ret;
}	

typedef Handler<CFolder> CFolderHandler;
typedef Handler<CProto> CProtoHandler;

// CFolders
// Stores the available Folders
// For Example: Inbox, Sent, MBox, etc

struct folder_info {
	gchar* name;
	CFolder* folder;
};

class CFolders {
	public:
		// CFolders::add
		// Name can be freed after calling this function
		// CFolder must not
		void add(CFolder* folder, gchar* name);
		void remove(gint); // maybe never, else TODO
		void remove(gchar* name) { remove(getfi(name)); }
		void remove(folder_info* data) {
			g_list_remove(folders, data); delete[] data;
		}
		// CFolder::get
		// Returns NULL if no Folder found
		CFolder* get(gchar* name) {
			folder_info *tmp = getfi(name);
			return tmp ? tmp->folder : NULL;
		}

		// callback should _not_ free anything it is given
		void enumerate(gboolean (callback)(CFolder*,gchar*)); 
		// ..obvious what following functions do...
		GtkWidget* makeCList(); // TODO
		GtkWidget* makeCombo(); // TODO
	protected:
		GList* folders;
		folder_info* getfi(gchar*);
};

#define CAN_CHANGE(a) \
	gtk_signal_connect(GTK_OBJECT(a), "changed", \
	GTK_SIGNAL_FUNC(entry_changed), prefs);


#endif //__MMC_PROTOCOLS_H
